#!/usr/bin/env python
"The rainbow server browser."

#    RN|Launcher, The rainbow server browser.
#    Copyright (C) 2008-2009  Joakim Soderlund
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import sys

## Attempts to find rnlauncher.glade, modules and data.
if os.path.exists(sys.path[0] + os.sep + "rnlauncher.glade"):
	glade_dir = sys.path[0] + os.sep
elif os.path.exists("/usr/share/rnlauncher/rnlauncher.glade"):
	glade_dir = "/usr/share/rnlauncher/"
	sys.path.append("/usr/share/rnlauncher/")
elif os.path.exists("/usr/local/share/rnlauncher/rnlauncher.glade"):
	glade_dir = "/usr/local/share/rnlauncher/"
	sys.path.append("/usr/local/share/rnlauncher/")
else:
	print("Error: Could not find rnlauncher.glade in\n - %s\n - /usr/share/rnlauncher/ \n - /usr/local/share/rnlauncher/" % glade_dir)
	exit()

import gtk
import gtk.glade
import pyquake3
import urllib
import configobj

from sys import exit
from Queue import Queue
from time import sleep
from xml.dom import minidom
from threading import Thread
from pyquake3master import fetchServerList

# Translation support.
try:
	gtk.glade.bindtextdomain("rnlauncher", glade_dir + "locale")
	gtk.glade.textdomain("rnlauncher")
except:
	print("Warning: Could not set locale!")

__author__       = "Joakim Soderlund"
__license__      = "GPLv2"
__copyright__    = "Copyright (C) 2008-2009  Joakim Soderlund"
__date__         = "2009-01-17"
__version__      = "1.6"

cfgCompVers      = "1.0"
app_name         = "rnlauncher"
app_version      = __version__
locale_dir       = ""
queryThreads     = 32

gtk.gdk.threads_init()

## The optional notifier.
try:
	import pynotify

	pynotify.init("RN|Launcher")
	
	def notify(body, title="RN|Launcher", time=5000, priority=pynotify.URGENCY_NORMAL):
		statusNote = pynotify.Notification(title, body, "dialog-info")
		statusNote.set_urgency(priority)
		statusNote.set_timeout(time)
		if StatusIcon.statusIcon.get_property("visible"):
			statusNote.attach_to_status_icon(StatusIcon.statusIcon)
		statusNote.show()
		
except:
	print("Warning: Could not find pynotify!")
	print("Will print notifications to the command line instead.")
	
	def notify(body, title="RN|Launcher", *args):
		print("%s: %s" % (title, body))


def loadConfig():
	"Loads the config file, returns a config Object."
	if os.name == "nt":
		config = configobj.ConfigObj("config.ini")
	else:
		if not os.path.isdir(os.path.join(os.path.expanduser("~"), ".rnlauncher")):
			os.mkdir(os.path.join(os.path.expanduser("~"), ".rnlauncher"))
		config = configobj.ConfigObj(os.path.join(os.path.expanduser("~"), ".rnlauncher", "config.ini"))

	try:
		if config["main"]["cfgVer"] != cfgCompVers:
			raise Exception("Invalid config version!")
	except:
		config["main"] = {"cfgVer":"1.0", "refreshOnStartup":0, "oa071bin":"", "oa077bin":""}
		config.write()

	if not config["main"].has_key("oa080bin"):
		config["main"]["oa080bin"] = ""
		config.write()
		
	if not config["main"].has_key("oa081bin"):
		config["main"]["oa081bin"] = ""
		config.write()

	if not "serverListSource" in config["main"]:
		config["main"]["serverListSource"] = "dpmaster.deathmask.net"
		config.write()

	if not "showStatusIcon" in config["main"]:
		config["main"]["showStatusIcon"] = "when required"
		config.write()

	if not "lastMainWindowSize" in config["main"]:
		config["main"]["lastMainWindowSize"] = (550, 290)
		config.write()
		
	if not "lastMainWindowPosition" in config["main"]:
		config["main"]["lastMainWindowPosition"] = (0, 0)
		config.write()
		
	if not "lastHPanedViewPosition" in config["main"]:
		config["main"]["lastHPanedViewPosition"] = 275
		config.write()
		
	if not "lastVPanedViewPosition" in config["main"]:
		config["main"]["lastVPanedViewPosition"] = 100
		config.write()
		
	return config

def threadedQuery():
	"Queries servers and adds them to a treeview, should be in a separate thread."
	while True:
		try:
			address, rowNum = mw.queue.get()
			server = pyquake3.PyQuake3(address)
			server.update()
			mw.rnServerDict[address] = server

			if server.vars["protocol"] == "69" and not os.path.isfile(config["main"]["oa077bin"]):
					pass
			elif server.vars["protocol"] == "68" and not os.path.isfile(config["main"]["oa071bin"]):
					pass
			elif server.vars["protocol"] == "70" and not os.path.isfile(config["main"]["oa080bin"]):
					pass
			elif server.vars["protocol"] == "71" and not os.path.isfile(config["main"]["oa081bin"]):
					pass
			else:
				if server.vars["protocol"] == "69":
					gameVer = "OA 0.7.7"
				elif  server.vars["protocol"] == "68":
					gameVer = "OA 0.7.1"
				elif  server.vars["protocol"] == "70":
					gameVer = "OA 0.8.0"
				elif server.vars["protocol"] == "71":
					gameVer = "OA 0.8.1"

				# Cleans up some server names.
				name = server.vars["sv_hostname"][:50]
				name = name.replace("  ", " ")
				name = name.strip()
				for num in range(9):
					name = name.replace("^%d" % num, "")
					name = name.replace("^%d" % num, "")
					name = name.strip()
				mw.rnServerDict[address].cleanName = name
				if name.startswith("RN|"):
					name = name.strip()
					name = name[3:]
					name = name.strip()
					name = "RN|" + name
					name = name.split("|")
					name = name[0] + "|" + name[1]
					name = name.strip()
					name = name.strip("071")
					name = name.strip("077")
					name = name.strip("080")
					name = name.strip("081")
					name = name.replace("-", "")
					name = name.replace("^", "")
					name = name.replace("*", " ")
					name = name.strip()
					mw.rnServerDict[address].cleanName = name
				if rowNum == "all":
					mw.serverList.append([name, server.vars["mapname"], "%02d/%02d" % (len(server.players), int(server.vars["sv_maxclients"])), "%03d" % server.ping, gameVer, address])
				else:
					mw.serverList[rowNum] = [name, server.vars["mapname"], "%02d/%02d" % (len(server.players), int(server.vars["sv_maxclients"])), "%03d" % server.ping, gameVer, address]
			if hasattr(mw.queue, "task_done"):
				mw.queue.task_done()
		except:
			if hasattr(mw.queue, "task_done"):
				mw.queue.task_done()

def getServerInfo():
	"Updates the server-list."
	mw.btn_refreshServers.set_sensitive(0)
	mw.btn_refreshSelected.set_sensitive(0)
	mw.btn_connect.set_sensitive(0)
	mw.btn_serverProperties.set_sensitive(0)
	mw.mi_refreshServers.set_sensitive(0)
	mw.mi_refreshSelected.set_sensitive(0)
	mw.mi_connect.set_sensitive(0)
	mw.mi_serverProperties.set_sensitive(0)
	mw.serverList.clear()
	mw.varsList.clear()
	mw.playerList.clear()
	if not os.path.isfile(config["main"]["oa071bin"]) and not os.path.isfile(config["main"]["oa077bin"]) and not os.path.isfile(config["main"]["oa080bin"]) and not os.path.isfile(config["main"]["oa081bin"]):
		mw.statusbar.push(0, "No OpenArena binaries found! Check your settings!")
		sleep(10)
		mw.btn_refreshServers.set_sensitive(1)
		mw.mi_refreshServers.set_sensitive(1)
		mw.statusbar.push(0, " ")
		return

	if config["main"]["serverListSource"].startswith("http://") and config["main"]["serverListSource"].endswith(".xml"):
		mw.statusbar.push(0, "Retrieving server list from furver.se...")
		xmldata = urllib.urlopen(config["main"]["serverListSource"]).read()
		xmldoc = minidom.parseString(xmldata)
		rnServerList = []		
		for server in xmldoc.getElementsByTagName("server"):
			if server.getElementsByTagName("type")[0].firstChild.toxml() == "quake3":
				rnServerList.append(	server.getElementsByTagName("address")[0].firstChild.toxml() + ":" +
									server.getElementsByTagName("port")[0].firstChild.toxml())
	else:
		mw.statusbar.push(0, "Retrieving server list from dpmaster.deathmask.net...")
		rnServerList = []
		for protocol in range(68, 72):
			try:
				rnServerList.extend(fetchServerList(config["main"]["serverListSource"], protocol=protocol, timeout=10))
			except:
				pass

	mw.statusbar.push(0, "Retrieving server info from %s servers..." % len(rnServerList))
	mw.rnServerDict = {}
	for address in rnServerList:
		mw.queue.put((address, "all"))
	if hasattr(mw.queue, "task_done"):
		mw.queue.join()
	else:
		sleep(10)
	mw.statusbar.push(0, " ")
	mw.btn_refreshServers.set_sensitive(1)
	mw.mi_refreshServers.set_sensitive(1)

def runMessages(serverIP=None):
	"Tells the user that we are launching OpenArena."
	if serverIP:
		serverName = mw.rnServerDict[serverIP].cleanName
	else:
		serverName = "server"
	mw.btn_connect.set_sensitive(0)
	mw.btn_serverProperties.set_sensitive(0)
	mw.mi_connect.set_sensitive(0)
	mw.mi_serverProperties.set_sensitive(0)
	mw.statusbar.push(0, "Connecting to %s..." % serverName)
	sleep(10)
	mw.statusbar.push(0, " ")
	mw.btn_connect.set_sensitive(1)
	mw.btn_serverProperties.set_sensitive(1)
	mw.mi_connect.set_sensitive(1)
	mw.mi_serverProperties.set_sensitive(1)

# We use a thread to start the threads because we don't want
# to slow down the start of the GUI... Dirty eh?
def startQueryThreads():
	"Starts query threads."
	for x in range(queryThreads):
		queryThread = Thread(target=threadedQuery)
		queryThread.setDaemon(True)
		queryThread.start()

def querySelectedServer():
	"Queries a server."
	mw.btn_refreshSelected.set_sensitive(0)
	mw.mi_refreshSelected.set_sensitive(0)
	serverIP, rowNum = mw.getSelectedIP(returnRowNum=True)
	mw.statusbar.push(0, "Updating %s..." % mw.rnServerDict[serverIP].cleanName)
	mw.queue.put((serverIP, rowNum))
	if hasattr(mw.queue, "task_done"):
		mw.queue.join()
	else:
		sleep(3)
	mw.updateLists(serverIP)
	mw.statusbar.push(0, "")
	mw.btn_refreshSelected.set_sensitive(1)
	mw.mi_refreshSelected.set_sensitive(1)
		
def connectToIP(serverIP):
	"Connects to a server using an IP-number."
	mw.btn_connect.set_sensitive(0)
	mw.btn_serverProperties.set_sensitive(0)
	mw.mi_connect.set_sensitive(0)
	mw.mi_serverProperties.set_sensitive(0)
	mw.statusbar.push(0, "Attempting to connect to %s..." % serverIP)
	mw.queue.put((serverIP, "all"))
	if hasattr(mw.queue, "task_done"):
		mw.queue.join()
	else:
		sleep(5)
	
	if not serverIP in mw.rnServerDict:
		mw.statusbar.push(0, "Could not connect to %s." % serverIP)
	else:
		if mw.rnServerDict[serverIP].vars["protocol"] == "69" and not os.path.isfile(config["main"]["oa077bin"]):
			mw.statusbar.push(0, "The server you specified uses OpenArena 0.7.7 which you do not have installed.")
		elif mw.rnServerDict[serverIP].vars["protocol"] == "68" and not os.path.isfile(config["main"]["oa071bin"]):
			mw.statusbar.push(0, "The server you specified uses OpenArena 0.7.1 which you do not have installed.")
		elif mw.rnServerDict[serverIP].vars["protocol"] == "70" and not os.path.isfile(config["main"]["oa080bin"]):
			mw.statusbar.push(0, "The server you specified uses OpenArena 0.8.0 which you do not have installed.")
		elif mw.rnServerDict[serverIP].vars["protocol"] == "71" and not os.path.isfile(config["main"]["oa081bin"]):
			mw.statusbar.push(0, "The server you specified uses OpenArena 0.8.1 which you do not have installed.")
		else:
			if mw.rnServerDict[serverIP].vars["protocol"] == "69":
				oabin = config["main"]["oa077bin"]
			elif mw.rnServerDict[serverIP].vars["protocol"] == "68":
				oabin = config["main"]["oa071bin"]
			elif mw.rnServerDict[serverIP].vars["protocol"] == "70":
				oabin = config["main"]["oa080bin"]
			elif mw.rnServerDict[serverIP].vars["protocol"] == "71":
				oabin = config["main"]["oa081bin"] 
		
			runopenarena = Thread(target=os.popen, args=("%s +connect \"%s\" " % (oabin, serverIP), ))
			mw.statusbar.push(0, "Connecting to %s..." % mw.rnServerDict[serverIP].cleanName)
			runopenarena.start()
	sleep(10)
	mw.statusbar.push(0, "")

class mainWindow():
	"RN|Launcher's main window."
	def __init__(self):
		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "mainWindow")

		signals={	"on_btn_refreshServers_clicked":self.on_btn_refreshServers_clicked,
					"on_btn_refreshSelected_clicked":self.on_btn_refreshSelected_clicked,
					"on_btn_connect_clicked":self.on_btn_connect_clicked,
					"on_btn_serverProperties_clicked":self.on_btn_serverProperties_clicked,
					"on_btn_mainSettings_clicked":self.on_btn_mainSettings_clicked,
					"on_serverView_row_activated":self.on_serverView_row_activated,
					"on_serverView_cursor_changed":self.on_serverView_cursor_changed,
					"on_mi_playersView_toggled":self.on_mi_playersView_toggled,
					"on_mi_varsView_toggled":self.on_mi_varsView_toggled,
					"on_mi_connectToIP_activate":self.on_mi_connectToIP_activate,
					"on_menu_about_activate":self.on_menu_about_activate,
					"on_mainWindow_destroy":self.on_mainWindow_destroy,
					"on_mainWindow_configure_event":self.configureEvent,
					"on_views_button_release_event":self.viewsConfigureEvent}

		self.wTree.signal_autoconnect(signals)

		global StatusIcon
		StatusIcon = statusIcon()

		self.mainWindow = self.wTree.get_widget("mainWindow")
		self.serverView = self.wTree.get_widget("serverView")
		self.playerView = self.wTree.get_widget("playerView")
		self.varsView = self.wTree.get_widget("varsView")
		self.hpanedView = self.wTree.get_widget("hpanedView")
		self.vpanedView = self.wTree.get_widget("vpanedView")
		self.statusbar = self.wTree.get_widget("statusbar")
		self.varsView = self.wTree.get_widget("varsView")
		self.btn_refreshServers = self.wTree.get_widget("btn_refreshServers")
		self.btn_refreshSelected = self.wTree.get_widget("btn_refreshSelected")
		self.btn_connect = self.wTree.get_widget("btn_connect")
		self.btn_serverProperties = self.wTree.get_widget("btn_serverProperties")
		self.mi_refreshServers = self.wTree.get_widget("mi_refreshServers")
		self.mi_refreshSelected = self.wTree.get_widget("mi_refreshSelected")
		self.mi_connect = self.wTree.get_widget("mi_connect")
		self.mi_serverProperties = self.wTree.get_widget("mi_serverProperties")
		self.mi_playersView = self.wTree.get_widget("mi_playersView")
		self.mi_varsView = self.wTree.get_widget("mi_varsView")
		self.playerViewWindow = self.wTree.get_widget("playerViewWindow")
		self.varsViewWindow = self.wTree.get_widget("varsViewWindow")

		self.mainWindow.resize(int(config["main"]["lastMainWindowSize"][0]), int(config["main"]["lastMainWindowSize"][1]))
		self.mainWindow.move(int(config["main"]["lastMainWindowPosition"][0]), int(config["main"]["lastMainWindowPosition"][1]))
		self.hpanedView.set_position(int(config["main"]["lastHPanedViewPosition"]))
		self.vpanedView.set_position(int(config["main"]["lastVPanedViewPosition"]))

		self.addViewColumn("Server name", 0, "serverView")
		self.addViewColumn("Map", 1, "serverView")
		self.addViewColumn("Players", 2, "serverView")
		self.addViewColumn("Ping", 3, "serverView")
		self.addViewColumn("Version", 4, "serverView")
		self.addViewColumn("Address", 5, "serverView")

		self.serverList = gtk.ListStore(str, str, str, str, str, str)
		self.serverView.set_model(self.serverList)

		self.addViewColumn("Player name", 0, "playerView")
		self.addViewColumn("Score", 1, "playerView")
		self.addViewColumn("Ping", 2, "playerView")

		self.playerList = gtk.ListStore(str, str, str)
		self.playerView.set_model(self.playerList)

		self.addViewColumn("Variable", 0, "varsView")
		self.addViewColumn("Value", 1, "varsView")

		self.varsList = gtk.ListStore(str, str)
		self.varsView.set_model(self.varsList)

		self.statusbar.push(0, " ")
		self.queue = Queue()
		self.mainWindow.show()

	def configureEvent(self, *args):
		"Called when the windows gets moved or resized."
		config["main"]["lastMainWindowSize"] = self.mainWindow.get_size()
		config["main"]["lastMainWindowPosition"] = self.mainWindow.get_position()
		config["main"]["lastHPanedViewPosition"] = self.hpanedView.get_position()
		config["main"]["lastVPanedViewPosition"] = self.vpanedView.get_position()
		
	def viewsConfigureEvent(self, *args):
		"Called when the views gets resized."
		config["main"]["lastHPanedViewPosition"] = self.hpanedView.get_position()
		config["main"]["lastVPanedViewPosition"] = self.vpanedView.get_position()

	def addViewColumn(self, colTitle, colID, view):
		"Creats columns for the server view."
		column = gtk.TreeViewColumn(colTitle, gtk.CellRendererText(), text=colID)
		column.set_resizable(True)
		column.set_sort_column_id(colID)
		if view == "serverView":
			self.serverView.append_column(column)
		elif view == "playerView":
			self.playerView.append_column(column)
		elif view == "varsView":
			self.varsView.append_column(column)
		
	def on_btn_refreshServers_clicked(self, widget, *args):
		"Called when refresh servers is clicked."
		self.refreshServers()

	def on_btn_refreshSelected_clicked(self, widget, *args):
		"Called when refresh selected is clicked."
		queryselectedserver = Thread(target=querySelectedServer)
		queryselectedserver.setDaemon(True)
		queryselectedserver.start()

	def refreshServers(self):
		"Refreshes the server list."
		getserverinfo = Thread(target=getServerInfo)
		getserverinfo.setDaemon(True)
		getserverinfo.start()

	def getSelectedIP(self, returnRowNum=False):
		"Returns the address of the selected server."
		rowNum = int("%s" % self.serverView.get_selection().get_selected_rows()[1][0])
		item = self.serverList.get_iter_first()
		cats = []
		while ( item != None ):
			cats.append(self.serverList.get_value (item, 5))
			item = self.serverList.iter_next(item)
		if returnRowNum:
			return cats[rowNum], rowNum
		else:
			return cats[rowNum]

	def connectToServer(self):
		"Connects to the selected server."
		serverIP = self.getSelectedIP()

		if self.rnServerDict[serverIP].vars["protocol"] == "69":
			oabin = config["main"]["oa077bin"]
		elif self.rnServerDict[serverIP].vars["protocol"] == "68":
			oabin = config["main"]["oa071bin"]
		elif self.rnServerDict[serverIP].vars["protocol"] == "70":
			oabin = config["main"]["oa080bin"]
		elif self.rnServerDict[serverIP].vars["protocol"] == "71":
			oabin = config["main"]["oa081bin"] 
		else:
			raise Exception("Error! Launched incompatible protocol version!")

		try:
			rconpassword = config[serverIP]["rconpassword"]
		except:
			rconpassword = ""
			
		try:
			joinpassword = config[serverIP]["joinpassword"]
		except:
			joinpassword = ""

		runopenarena = Thread(target=os.popen, args=("%s +connect \"%s\" +set rconpassword \"%s\" +set password \"%s\"" % (oabin, serverIP, rconpassword, joinpassword), ))
		runmessages = Thread(target=runMessages, args=(serverIP, ))
		runmessages.setDaemon(True)
		runmessages.start()
		runopenarena.start()


	def on_btn_connect_clicked(self, widget, *args):
		"Connects to a server."
		self.connectToServer()

	def on_btn_serverProperties_clicked(self, widget, *args):
		"Handles rcon passwords and private server passwords."
		serverIP = self.getSelectedIP()
		self.serverProperties = serverSettings()
		
		try:
			rconpassword = config[serverIP]["rconpassword"]
		except:
			rconpassword = ""
			
		try:
			joinpassword = config[serverIP]["joinpassword"]
		except:
			joinpassword = ""
			
		self.serverProperties.recieveServerInfo(serverIP, self.rnServerDict[serverIP], rconpassword, joinpassword)

	def on_btn_mainSettings_clicked(self, widget, *args):
		"Opens a new main settings dialog."
		self.mainSettings = mainSettings()

	def on_serverView_row_activated(self, widget, *args):
		"""Connects to a server when a row in the main treeview has been activated.
		This is usally done by double clicking on a server or hitting return or space
		when a server is selected."""
		self.connectToServer()

	def on_serverView_cursor_changed(self, widget, *args):
		"Called when a server has been selected."
		mw.btn_refreshSelected.set_sensitive(1)
		mw.btn_connect.set_sensitive(1)
		mw.btn_serverProperties.set_sensitive(1)
		mw.mi_refreshSelected.set_sensitive(1)
		mw.mi_connect.set_sensitive(1)
		mw.mi_serverProperties.set_sensitive(1)
		self.updateLists(self.getSelectedIP())

	def updateLists(self, serverIP):
		"Updates the player list and vars list."
		self.playerList.clear()
		self.varsList.clear()
		for player in self.rnServerDict[serverIP].players:
			player.name_clean = player.name
			# Removes the colors from the player colors
			for integer in range(9):
				player.name_clean = player.name_clean.replace("^%d" % integer, "")
			self.playerList.append([player.name_clean, "%03d" % int(player.frags), "%03d" % int(player.ping)])
		for k, v in self.rnServerDict[serverIP].vars.items():
			self.varsList.append([k, v])

	def on_mi_playersView_toggled(self, widget, *args):
		"Shows or hides the player view."
		if self.mi_playersView.get_active():
			self.playerViewWindow.show()
		else:
			self.playerViewWindow.hide()

	def on_mi_varsView_toggled(self, widget, *args):
		"Shows or hides the vars view."
		if self.mi_varsView.get_active():
			self.varsViewWindow.show()
		else:
			self.varsViewWindow.hide()

	def on_menu_about_activate(self, widget, *args):
		"Shows the about dialog."
		aboutdialog = aboutDialog()
		
	def on_mi_connectToIP_activate(self, widget, *args):
		"Opens a new connect to IP dialog."
		connecttoip = ipDialog()

	def on_mainWindow_destroy(self, widget, *args):
		"Destroys the main window."
		gtk.main_quit()

	def run(self):
		"Initializes GTK."
		if config["main"]["showStatusIcon"] == "always":
			StatusIcon.show()
		gtk.gdk.threads_enter()
		gtk.main()
		gtk.gdk.threads_leave()

class aboutDialog():
	"RN|Launcher's about dialog."
	def __init__(self):
		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "aboutDialog")
		signals = {"on_aboutDialog_destroy":self.on_aboutDialog_destroy}
		self.wTree.signal_autoconnect(signals)
		self.aboutDialog = self.wTree.get_widget("aboutDialog")
		self.aboutDialog.set_version(__version__)

	def on_aboutDialog_destroy(self, widget, *args):
		"Destroys the dialog when we close it."
		self.aboutDialog.destroy()

class mainSettings():
	"RN|Launcher's main settings dialog."
	def __init__(self):

		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "mainSettings")

		signals = {	"on_btn_ok2_clicked":self.on_btn_ok2_clicked,
					"on_btn_oa071bin_clicked":self.on_btn_oa071bin_clicked,
					"on_btn_oa077bin_clicked":self.on_btn_oa077bin_clicked,
					"on_btn_oa080bin_clicked":self.on_btn_oa080bin_clicked,
					"on_btn_oa081bin_clicked":self.on_btn_oa081bin_clicked,
					"on_btn_abort2_clicked":self.on_btn_abort2_clicked}

		self.wTree.signal_autoconnect(signals)

		self.mainSettings = self.wTree.get_widget("mainSettings")

		self.ent_oa071bin = self.wTree.get_widget("ent_oa071bin")
		self.ent_oa077bin = self.wTree.get_widget("ent_oa077bin")
		self.ent_oa080bin = self.wTree.get_widget("ent_oa080bin")
		self.ent_oa081bin = self.wTree.get_widget("ent_oa081bin")
		self.cmb_statusIcon = self.wTree.get_widget("cmb_statusIcon")
		self.cbx_refreshOnStartup = self.wTree.get_widget("cbx_refreshOnStartup")
		self.cbx_showPlayersOnStartup = self.wTree.get_widget("cbx_showPlayersOnStartup")
		self.cbx_showVarsOnStartup = self.wTree.get_widget("cbx_showVarsOnStartup")
		self.cbx_showRainbowServersOnly = self.wTree.get_widget("cbx_showRainbowServersOnly")

		self.ent_oa071bin.set_text(config["main"]["oa071bin"])
		self.ent_oa077bin.set_text(config["main"]["oa077bin"])
		self.ent_oa080bin.set_text(config["main"]["oa080bin"])
		self.ent_oa081bin.set_text(config["main"]["oa081bin"])
		self.cbx_refreshOnStartup.set_active(int(config["main"]["refreshOnStartup"]))
		self.cbx_showPlayersOnStartup.set_active(int(config["main"]["showPlayersOnStartup"]))
		self.cbx_showVarsOnStartup.set_active(int(config["main"]["showVarsOnStartup"]))
		
		## Sets the active item in the tray icon combo box.
		self.cmb_statusIcon.set_active(["always", "when required", "never"].index(config["main"]["showStatusIcon"]))

		if config["main"]["serverListSource"].startswith("http://") and config["main"]["serverListSource"].endswith(".xml"):
			self.cbx_showRainbowServersOnly.set_active(1)

	def on_btn_ok2_clicked(self, widget, *args):
		"Called when ok is clicked, saves settings."
		config["main"]["oa071bin"] = self.ent_oa071bin.get_text()
		config["main"]["oa077bin"] = self.ent_oa077bin.get_text()
		config["main"]["oa080bin"] = self.ent_oa080bin.get_text()
		config["main"]["oa081bin"] = self.ent_oa081bin.get_text()
		if self.cbx_refreshOnStartup.get_active():
			config["main"]["refreshOnStartup"] = 1
		else:
			config["main"]["refreshOnStartup"] = 0

		if self.cbx_showPlayersOnStartup.get_active():
			config["main"]["showPlayersOnStartup"] = 1
		else:
			config["main"]["showPlayersOnStartup"] = 0

		if self.cbx_showVarsOnStartup.get_active():
			config["main"]["showVarsOnStartup"] = 1
		else:
			config["main"]["showVarsOnStartup"] = 0

		if self.cbx_showRainbowServersOnly.get_active():
			config["main"]["serverListSource"] = "http://rainbow.furver.se/rainbow_server_list.xml"
		else:
			config["main"]["serverListSource"] = "dpmaster.deathmask.net"
			
		config["main"]["showStatusIcon"] = self.cmb_statusIcon.get_active_text()
		if config["main"]["showStatusIcon"] == "always":
			StatusIcon.show()
		elif config["main"]["showStatusIcon"] == "when required":
			if StatusIcon.users == 0:
				StatusIcon.hide()
			else:
				StatusIcon.show()
		else:
			StatusIcon.hide()

		config.write()
		self.mainSettings.destroy()

	def on_btn_oa071bin_clicked(self, widget, *args):
		"Brings up a file chooser dialog for OpenArena 0.7.1."
		self.fileChooser = fileChooser(self.ent_oa071bin.set_text)
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.7.1 binary")

	def on_btn_oa077bin_clicked(self, widget, *args):
		"Brings up a file chooser dialog for OpenArena 0.7.7."
		self.fileChooser = fileChooser(self.ent_oa077bin.set_text)
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.7.7 binary")

	def on_btn_oa080bin_clicked(self, widget, *args):
		"Brings up a file chooser dialog for OpenArena 0.8.0."
		self.fileChooser = fileChooser(self.ent_oa080bin.set_text)
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.8.0 binary")
		
	def on_btn_oa081bin_clicked(self, widget, *args):
		"Brings up a file chooser dialog for OpenArena 0.8.1."
		self.fileChooser = fileChooser(self.ent_oa081bin.set_text)
		self.fileChooser.fileChooser.set_title("Select your Open Arena 0.8.1 binary")

	def on_btn_abort2_clicked(self, widget, *args):
		"Destroys the dialog without saving anything."
		self.mainSettings.destroy()

class serverSettings():
	"RN|Launcher's server properties dialog."
	def __init__(self):

		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "serverSettings")

		signals = {	"on_btn_ok_clicked":self.on_btn_ok_clicked,
					"on_btn_abort_clicked":self.on_btn_abort_clicked}

		self.wTree.signal_autoconnect(signals)

		self.serverSettings = self.wTree.get_widget("serverSettings")
		self.lab_serverName = self.wTree.get_widget("lab_serverName")
		self.ent_rconPassword = self.wTree.get_widget("ent_rconPassword")
		self.ent_joinPassword = self.wTree.get_widget("ent_joinPassword")

	def recieveServerInfo(self, serverIP, serverData, rconpassword, joinpassword):
		"Adds data to the dialog."
		self.serverIP = serverIP
		self.serverData = serverData
		self.serverSettings.set_title(serverIP)
		self.ent_rconPassword.set_text(rconpassword)
		self.ent_joinPassword.set_text(joinpassword)

	def on_btn_ok_clicked(self, widget, *args):
		"Destroys the dialog and saves settings."
		try:
			config[self.serverIP]["rconpassword"] = self.ent_rconPassword.get_text()
		except KeyError:
			config[self.serverIP] = {}
			config[self.serverIP]["rconpassword"] = self.ent_rconPassword.get_text()
			
		try:
			config[self.serverIP]["joinpassword"] = self.ent_joinPassword.get_text()
		except KeyError:
			config[self.serverIP] = {}
			config[self.serverIP]["joinpassword"] = self.ent_joinPassword.get_text()
		
		config.write()
		self.serverSettings.destroy()
		
	def on_btn_abort_clicked(self, widget, *args):
		"Destroys the dialog without saving settings."
		self.serverSettings.destroy()

class fileChooser():
	"RN|Launcher's file chooser dialog."
	def __init__(self, reporter):
		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "fileChooser")

		signals = {	"on_btn_okfc_clicked":self.on_btn_okfc_clicked,
					"on_btn_abortfc_clicked":self.on_btn_abortfc_clicked}

		self.wTree.signal_autoconnect(signals)

		self.fileChooser = self.wTree.get_widget("fileChooser")
		self.reporter = reporter

	def on_btn_okfc_clicked(self, widget, *args):
		"Destroys the dialog and sends data to the report hook."
		self.reporter(self.fileChooser.get_filename())
		self.fileChooser.destroy()

	def on_btn_abortfc_clicked(self, widget, *args):
		"Destroys the dialog."
		self.fileChooser.destroy()

class statusIcon():
	"RN|Launcher's status (tray) icon."
	def __init__(self):
		self.statusIcon = gtk.StatusIcon()
		self.statusIcon.set_from_file(glade_dir + "rnlogo.png")
		self.statusIcon.connect("activate", self.on_statusIcon_clicked)
		self.statusIcon.connect('popup-menu', self.on_statusIcon_rightClicked)
		self.statusIcon.set_visible(False)
				
		self.smTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "statusMenu")
		
		self.statusMenu = self.smTree.get_widget("statusMenu")
		
		self.smTree.signal_autoconnect({"on_smb_quit_activate":gtk.main_quit})
		
		self.users = 0
		
	def on_statusIcon_clicked(self, *args):
		"Hides the main window if the icon has been clicked."
		if mw.mainWindow.get_property("visible"):
			mw.mainWindow.hide()
		else:
			mw.mainWindow.show()
			
	def on_statusIcon_rightClicked(self, *args):
		"Shows a menu if the icon has been right-clicked."
		self.statusMenu.show_all()
		self.statusMenu.popup(None, None, None, args[1], args[2])
			
	def show(self):
		"Shows the icon."
		self.statusIcon.set_visible(True)

	def hide(self):
		"Hides the icon."
		self.statusIcon.set_visible(False)
		
class ipDialog():
	"RN|Launcher's connect to IP dialog."
	def __init__(self):
		self.wTree = gtk.glade.XML(glade_dir + "rnlauncher.glade", "ipDialog")
		signals = {	"on_btn_connect_to_ip_clicked":self.on_btn_connect_to_ip_clicked,
					"on_btn_abort_ip_clicked":self.on_btn_abort_ip_clicked}
		self.wTree.signal_autoconnect(signals)
		self.ipDialog = self.wTree.get_widget("ipDialog")
		self.ent_ip = self.wTree.get_widget("ent_ip")
		self.ent_port = self.wTree.get_widget("ent_port")
		self.btn_connect_to_ip = self.wTree.get_widget("btn_connect_to_ip")

	def on_btn_connect_to_ip_clicked (self, widget, *args):
		"Calls connectToIP."
		self.connectToIP()
	
	def on_btn_abort_ip_clicked (self, widget, *args):
		"Destroys the dialog wihout doing anything."
		self.ipDialog.destroy()		

	def connectToIP(self):
		"Connects to the selected server."
		self.ipDialog.destroy()
		ip = self.ent_ip.get_text()
		port = self.ent_port.get_text()
		serverIP = "%s:%s" % (ip, port)
		
		connect = Thread(target=connectToIP, args=[serverIP])
		connect.start()

if __name__ == "__main__":
	# The config object.
	config = loadConfig()
	
	# Creates a new main window.
	mw = mainWindow()
	
	if int(config["main"]["refreshOnStartup"]):
		mw.refreshServers()

	try:
		if int(config["main"]["showPlayersOnStartup"]):
			mw.mi_playersView.set_active(1)
	except:
		config["main"]["showPlayersOnStartup"] = 0

	try:
		if int(config["main"]["showVarsOnStartup"]):
			mw.mi_varsView.set_active(1)
	except:
			config["main"]["showVarsOnStartup"] = 0
	
	# Starts query threads.
	threadStarter = Thread(target=startQueryThreads)
	threadStarter.setDaemon(True)
	threadStarter.start()
	
	# Enters the gtk main loop.
	mw.run()
	
	# Writes to the config file and exits.
	config.write()
